﻿// Copyright (c) 2014 Nora
// Released under the MIT license
// http://opensource.org/licenses/mit-license.php
#include "../Header/OVRMirror.h"
#include "Common/Thread.hpp"

#ifdef _WIN32
#include "Windows/Window.hpp"
#include "Windows/Tasktray.hpp"
#include "Windows/OpenGL/OGLRenderer.hpp"
#include "Windows/DirectX9/DX9Renderer.hpp"
#include "Windows/DirectX11/DX11Renderer.hpp"
#endif

using namespace OVRMirror;
#ifdef _WIN32
using namespace OVRMirror::Windows;
#endif

static RecursiveMutex	_mutex;
static Window			_window;
#ifdef _WIN32
static Tasktray			_tasktray;
#endif
static int32_t			_deviceType = -1;
static IRenderer*		_renderer = nullptr;
static RendererParams	_rendererParams;
static bool				_isSetRenderEventID = false;
static int32_t			_renderEventID = 0;
static int32_t			_windowFlags = 0;

//--------------------------------------------------------------------------------------------------------------------------------------------------------

#ifdef _WIN32 // {

#pragma comment(lib, "winmm.lib")

struct Win32ThreadBoost
{
	Win32ThreadBoost() { ::timeBeginPeriod(1); }
	~Win32ThreadBoost() { ::timeEndPeriod(1); }
};

static Win32ThreadBoost _threadBoost;

#endif // } _WIN32

//--------------------------------------------------------------------------------------------------------------------------------------------------------

void UnitySetGraphicsDevice( void* device, int32_t deviceType, int32_t eventType )
{
	{
		LockGuardRecursiveMutex lock(_mutex);
		if( _deviceType != deviceType && _deviceType != -1 ) {
			if( _renderer != nullptr ) {
				delete _renderer;
				_renderer = nullptr;
			}
		}

		_deviceType = deviceType;

		if( _renderer == nullptr ) {
#ifdef _WIN32 // {
			if( deviceType == kGfxRendererOpenGL ) {
				_renderer = new OGLRenderer();
			}
			if( deviceType == kGfxRendererD3D9 ) {
				_renderer = new DX9Renderer();
			}
			if( deviceType == kGfxRendererD3D11 ) {
				_renderer = new DX11Renderer();
			}
#endif // } _WIN32
		}

		_window.SetDeviceType( deviceType );

		_rendererParams.device = device;
		_rendererParams.window = &_window;

		switch( eventType ) {
		case kGfxDeviceEventInitialize:
		case kGfxDeviceEventAfterReset:
			_rendererParams.deviceActivated = true;
			break;
		}
	}

	if( _renderer != nullptr ) { // Memo: Not require _mutex.lock()
		_renderer->SetGraphicsDevice( device, eventType );
	}

	{
		LockGuardRecursiveMutex lock(_mutex);
		switch( eventType ) {
		case kGfxDeviceEventShutdown:
		case kGfxDeviceEventBeforeReset:
			_rendererParams.device = 0;
			_rendererParams.deviceActivated = false;
			break;
		}
		if( eventType == kGfxDeviceEventShutdown ) {
			if( _renderer != nullptr ) {
				delete _renderer;
				_renderer = nullptr;
			}

			_window._DestroyWindow();
		}
	}
}

//--------------------------------------------------------------------------------------------------------------------------------------------------------

void UnityRenderEvent( int32_t eventID )
{
	bool isSetRenderEventID = false;
	int32_t renderEventID = 0;
	RendererParams rendererParams;

	{
		LockGuardRecursiveMutex lock(_mutex);
		isSetRenderEventID = _isSetRenderEventID;
		renderEventID = _renderEventID;
		rendererParams = _rendererParams;
	}

	if( !isSetRenderEventID || renderEventID == eventID ) {
		if( _renderer != nullptr ) { // Memo: Not require _mutex.lock()
			_renderer->Render( rendererParams );
			{
				LockGuardRecursiveMutex lock(_mutex);
				_rendererParams.ClearTrigger();
			}
		}
	}
}

//--------------------------------------------------------------------------------------------------------------------------------------------------------

void OVRMirrorCreate()
{
	LockGuardRecursiveMutex lock(_mutex);
#ifdef _WIN32
	_tasktray._CreateWindow();
#endif
}

void OVRMirrorDestroy()
{
	LockGuardRecursiveMutex lock(_mutex);
	_window._DestroyWindow();
	_rendererParams.openWindowTrigger = 0;
#ifdef _WIN32
	_tasktray._DestroyWindow();
#endif
}

void OVRMirrorSetWindowFlags( int32_t windowFlags )
{
	_windowFlags = windowFlags;
}

void OVRMirrorOpenWindow()
{
	LockGuardRecursiveMutex lock(_mutex);
	if( !_window.CheckHWnd() ) {
		const bool isFullscreen = (_rendererParams.fullscreenTrigger == 1);
		const bool isMaximized = (_windowFlags & 0x01) || (_rendererParams.maximizeTrigger == 1);
		_window._CreateWindow( _deviceType, _rendererParams.clientSize, isFullscreen, isMaximized );
		_rendererParams.openWindowTrigger = 1;
	}
#ifdef _WIN32
	_tasktray._CreateWindow();
#endif
}

void OVRMirrorCloseWindow()
{
	LockGuardRecursiveMutex lock(_mutex);
	_window._DestroyWindow();
	_rendererParams.openWindowTrigger = 0;
}

int32_t OVRMirrorGetWindowState()
{
	int32_t r = 0;
	{
		LockGuardRecursiveMutex lock(_mutex);
		r = _window.GetWindowStateImmediately();
	}

	return r;
}

uint64_t OVRMirrorGetClientSize( int32_t* clientSize )
{
	ISIZE32 r = { 0, 0 };
	{
		LockGuardRecursiveMutex lock(_mutex);
		r = _window.GetClientSizeImmediately();
	}

	return (uint64_t)r.w | ((uint64_t)r.h << 32u);
}

int32_t OVRMirrorSetRenderTexture( void* nativeTex )
{
	LockGuardRecursiveMutex lock(_mutex);
	if( _renderer != nullptr ) {
		_renderer->SetRenderTextureImmediately( nativeTex );
		return 1;
	}

	return 0;
}

void OVRMirrorSetViewportRect(  float x, float y, float w, float h  )
{
	LockGuardRecursiveMutex lock(_mutex);
	_rendererParams.viewportRect.x = x;
	_rendererParams.viewportRect.y = y;
	_rendererParams.viewportRect.w = w;
	_rendererParams.viewportRect.h = h;
}

void OVRMirrorSetRenderEventID( int32_t eventID )
{
	LockGuardRecursiveMutex lock(_mutex);
	_isSetRenderEventID = true;
	_renderEventID = eventID;
}

void OVRMirrorResetRenderEventID()
{
	LockGuardRecursiveMutex lock(_mutex);
	_isSetRenderEventID = false;
	_renderEventID = 0;
}
