// Copyright (c) 2014 Nora
// Released under the MIT license
// http://opensource.org/licenses/mit-license.php
#include "DX11Renderer.hpp"
#include "../Window.hpp"

namespace OVRMirror
{
namespace Windows
{

namespace {

#include "VertexShader.h"
#include "PixelShader.h"

} // namespace

//----------------------------------------------------------------------------------------------------------------------------------------------------

DX11Renderer::DX11Renderer()
	: _isDeviceInitialized( false )
	, _isDirtyRenderTexture( false )

	, _renderTexture( NULL )
	, _textureView( NULL )

	, _vertexBuffer( NULL )
	, _vertexShader( NULL )
	, _pixelShader( NULL )
	, _inputLayout( NULL )
	, _rasterState( NULL )
	, _blendState( NULL )
	, _depthState( NULL )
	, _samplerState( NULL )
	, _swapChain( NULL )
	, _backBuffer( NULL )
	, _backBufferRT( NULL )
{
	memset( &_backBufferSize, 0, sizeof(_backBufferSize) );
	memset( _backupViewports, 0, sizeof(_backupViewports) );
	memset( _backupVertexClassInstances, 0, sizeof(_backupVertexClassInstances) );
	memset( _backupPixelClassInstances, 0, sizeof(_backupPixelClassInstances) );
}

DX11Renderer::~DX11Renderer()
{
	_ShutdownDevice();
}

//----------------------------------------------------------------------------------------------------------------------------------------------------

void DX11Renderer::SetGraphicsDevice( void* device, int32_t eventType )
{
	switch( eventType ) {
	case kGfxDeviceEventInitialize:
	case kGfxDeviceEventAfterReset:
		// Nothing.
		break;
	case kGfxDeviceEventShutdown:
	case kGfxDeviceEventBeforeReset:
		_ShutdownDevice();
		break;
	}
}

void DX11Renderer::SetRenderTextureImmediately( void* renderTexture )
{
	LockGuardRecursiveMutex lock(_mutex);
	if( _renderTexture == renderTexture ) {
		return;
	}

	SAFE_RELEASE( _renderTexture );
	_renderTexture = static_cast< ID3D11Resource* >( renderTexture );
	_isDirtyRenderTexture = true;
	if( _renderTexture != NULL ) {
		_renderTexture->AddRef();
	}
}

void DX11Renderer::Render( const RendererParams& rendererParams )
{
	Window* window = static_cast< Window* >( rendererParams.window );
	if( window == nullptr ) {
		return;
	}

	bool isWindow = false;
	bool isMinimized = false;

	{
		LockGuardRecursiveMutex lock(window->mutex);
		isWindow = window->UnsafeCheckHWnd();
		isMinimized = window->UnsafeIsMinimized();
	}

	if( !isWindow || rendererParams.openWindowTrigger == 0 ) {
		_ShutdownDevice();
		return;
	}

	if( isMinimized ) {
		return;
	}

	if( rendererParams.device == NULL ) {
		return;
	}

	if( !_PrepareDevice( rendererParams ) ) {
		return;
	}

	ID3D11Device* device = static_cast< ID3D11Device* >( rendererParams.device );

	ID3D11DeviceContext* ctx = NULL;
	device->GetImmediateContext( &ctx );
	if( ctx == NULL ) {
		return;
	}

	{
		LockGuardRecursiveMutex lock(_mutex);
		if( _renderTexture != NULL && _isDirtyRenderTexture ) {
			_isDirtyRenderTexture = false;
			D3D11_RESOURCE_DIMENSION resourceDimension = D3D11_RESOURCE_DIMENSION_UNKNOWN;
			_renderTexture->GetType( &resourceDimension );
			if( resourceDimension == D3D11_RESOURCE_DIMENSION_TEXTURE2D ) {
				ID3D11Texture2D* texture = static_cast<ID3D11Texture2D*>( _renderTexture );
				D3D11_SHADER_RESOURCE_VIEW_DESC srvDesc;
				memset( &srvDesc, 0, sizeof(srvDesc) );
				D3D11_TEXTURE2D_DESC desc;
				memset( &desc, 0, sizeof(desc) );
				texture->GetDesc( &desc );
				//srvDesc.Format = desc.Format;
				srvDesc.Format = DXGI_FORMAT_R8G8B8A8_UNORM;
				srvDesc.ViewDimension = D3D11_SRV_DIMENSION_TEXTURE2D;
				//srvDesc.ViewDimension = D3D11_SRV_DIMENSION_TEXTURE2DMS;
				srvDesc.Texture2D.MostDetailedMip = 0;
				srvDesc.Texture2D.MipLevels = desc.MipLevels;
				device->CreateShaderResourceView( texture, &srvDesc, &_textureView ); 
			}
		}
	}

	if( _backBufferRT != NULL && _textureView != NULL && _vertexShader != NULL && _pixelShader != NULL ) {
		ID3D11RenderTargetView* backupRTV = NULL;
		ID3D11DepthStencilView* backupDSV = NULL;
		ctx->OMGetRenderTargets( 1, &backupRTV, &backupDSV );
		ctx->OMSetRenderTargets( 1, &_backBufferRT, NULL );
		float clearColor[4] = { 0.0f, 0.0f, 0.0f, 0.0f };
		ctx->ClearRenderTargetView( _backBufferRT, clearColor );

		ID3D11DepthStencilState* backupDepthState = NULL;
		UINT backupStencilRef = 0;
		ctx->OMGetDepthStencilState( &backupDepthState, &backupStencilRef );
		ctx->OMSetDepthStencilState( _depthState, 0 );

		ID3D11RasterizerState* backupRasterState = NULL;
		ctx->RSGetState( &backupRasterState );
		ctx->RSSetState( _rasterState );

		ID3D11BlendState* backupBlendState = NULL;
		float backupBlendFactor[4] = { 0, 0, 0, 0 };
		UINT backupSampleMask = 0;
		ctx->OMGetBlendState( &backupBlendState, backupBlendFactor, &backupSampleMask );
		ctx->OMSetBlendState( _blendState, NULL, 0xFFFFFFFF );

		UINT backupNumViewport = 0;
		ctx->RSGetViewports( &backupNumViewport, _backupViewports );

		{
			D3D11_VIEWPORT vp;
			vp.Width = (float)_backBufferSize.w;
			vp.Height = (float)_backBufferSize.h;
			vp.MinDepth = 0.0f;
			vp.MaxDepth = 1.0f;
			vp.TopLeftX = 0;
			vp.TopLeftY = 0;
			ctx->RSSetViewports( 1, &vp );
		}

		ID3D11InputLayout* backupInputLayout = NULL;
		ctx->IAGetInputLayout( &backupInputLayout );
		ctx->IASetInputLayout( _inputLayout );

		UINT backupVertexStride = 0;
		UINT backupVertexOffset = 0;
		ID3D11Buffer* backupVertexBuffer = NULL;
		ctx->IAGetVertexBuffers( 0, 1, &backupVertexBuffer, &backupVertexStride, &backupVertexOffset );

		{
			UINT stride = sizeof(RenderTexturePrimitiveCache::Vertex);
			UINT offset = 0;
			ctx->IASetVertexBuffers( 0, 1, &_vertexBuffer, &stride, &offset );
		}

		D3D11_PRIMITIVE_TOPOLOGY backupTopology = (D3D11_PRIMITIVE_TOPOLOGY)0;
		ctx->IAGetPrimitiveTopology( &backupTopology );
		ctx->IASetPrimitiveTopology( D3D11_PRIMITIVE_TOPOLOGY_TRIANGLESTRIP );

		ID3D11SamplerState* backupSamplerState = NULL;
		ctx->PSGetSamplers( 0, 1, &backupSamplerState );

		ID3D11ShaderResourceView* backupShaderResourceView = NULL;
		ctx->PSGetShaderResources( 0, 1, &backupShaderResourceView );

		// Backup VertexShader.
		ID3D11VertexShader* backupVertexShader = NULL;
		UINT backupVertexNumClassInstances = 0;
		ctx->VSGetShader( &backupVertexShader, _backupVertexClassInstances, &backupVertexNumClassInstances );

		// Backup PixelShader.
		ID3D11PixelShader* backupPixelShader = NULL;
		UINT backupPixelNumClassInstances = 0;
		ctx->PSGetShader( &backupPixelShader, _backupPixelClassInstances, &backupPixelNumClassInstances );

		ctx->VSSetShader( _vertexShader, NULL, 0 );
		ctx->PSSetShader( _pixelShader, NULL, 0 );
		ctx->PSSetSamplers( 0, 1, &_samplerState );
		ctx->PSSetShaderResources( 0, 1, &_textureView );

		ctx->Draw( 4, 0 );

		ctx->VSSetShader( backupVertexShader, _backupVertexClassInstances, backupVertexNumClassInstances );
		for( UINT i = 0; i != backupVertexNumClassInstances; ++i ) {
			SAFE_RELEASE( _backupVertexClassInstances[i] );
		}
		SAFE_RELEASE( backupVertexShader );

		ctx->PSSetShader( backupPixelShader, _backupPixelClassInstances, backupPixelNumClassInstances );
		for( UINT i = 0; i != backupPixelNumClassInstances; ++i ) {
			SAFE_RELEASE( _backupPixelClassInstances[i] );
		}
		SAFE_RELEASE( backupPixelShader );

		ctx->PSSetSamplers( 0, 1, &backupSamplerState );
		SAFE_RELEASE( backupSamplerState );

		ctx->PSSetShaderResources( 0, 1, &backupShaderResourceView );
		SAFE_RELEASE( backupShaderResourceView );

		ctx->IASetVertexBuffers( 0, 1, &backupVertexBuffer, &backupVertexStride, &backupVertexOffset );
		SAFE_RELEASE( backupVertexBuffer );

		ctx->IASetInputLayout( backupInputLayout );
		SAFE_RELEASE( backupInputLayout );

		ctx->IASetPrimitiveTopology( backupTopology );

		ctx->RSSetViewports( backupNumViewport, _backupViewports );

		ctx->OMSetBlendState( backupBlendState, backupBlendFactor, backupSampleMask );
		SAFE_RELEASE( backupBlendState );

		ctx->RSSetState( backupRasterState );
		SAFE_RELEASE( backupRasterState );

		ctx->OMSetDepthStencilState( backupDepthState, backupStencilRef );
		SAFE_RELEASE( backupDepthState );

		ctx->OMSetRenderTargets( 1, &backupRTV, backupDSV );
		SAFE_RELEASE( backupRTV );
		SAFE_RELEASE( backupDSV );
	}

	if( _swapChain != NULL ) {
		_swapChain->Present( 0, 0 );
	}

	ctx->Release();
}

bool DX11Renderer::_PrepareDevice( const RendererParams& rendererParams )
{
	Window* window = static_cast< Window* >( rendererParams.window );
	if( window == nullptr ) {
		return false;
	}

	ID3D11Device* device = static_cast< ID3D11Device* >( rendererParams.device );
	if( device == NULL ) {
		return false;
	}

	ISIZE32 clientSize = window->GetClientSizeImmediately();
	if( clientSize.w <= 0 || clientSize.h <= 0 ) {
		return false;
	}

	const FRECT32& viewportRect = rendererParams.viewportRect;
	if( viewportRect.w <= 0.0f || viewportRect.h <= 0.0f ) {
		return false;
	}
	
	{
		DXGI_SWAP_CHAIN_DESC scDesc;
		memset(&scDesc, 0, sizeof(scDesc));
		if( _swapChain != NULL ) {
			if( SUCCEEDED( _swapChain->GetDesc( &scDesc ) ) ) {
				HWND hWnd = scDesc.OutputWindow;
				bool isResetSwapChain = false;
				if( rendererParams.openWindowTrigger != -1 || hWnd != window->GetHWnd() || ::IsWindow( hWnd ) == FALSE ) {
					isResetSwapChain = true;
				}
				if( rendererParams.backBufferSize.w <= 0 ||
					rendererParams.backBufferSize.h <= 0 ) {
					UINT backBufferWidth = (UINT)scDesc.BufferDesc.Width;
					UINT backBufferHeight = (UINT)scDesc.BufferDesc.Height;
					if( backBufferWidth != clientSize.w || backBufferHeight != clientSize.h ) {
						isResetSwapChain = true;
					}
				}
				if( isResetSwapChain ) {
					SAFE_RELEASE( _backBufferRT );
					SAFE_RELEASE( _backBuffer );
					SAFE_RELEASE( _swapChain );
					memset( &_backBufferSize, 0, sizeof(_backBufferSize) );
					_isDeviceInitialized = false;
				}
			}
		}
	}

	if( !_isDeviceInitialized ) {
		_isDeviceInitialized = true;

		if( _samplerState == NULL ) {
			D3D11_SAMPLER_DESC samplerDesc;
			memset( &samplerDesc, 0, sizeof(samplerDesc) );
			samplerDesc.AddressU = D3D11_TEXTURE_ADDRESS_CLAMP;
			samplerDesc.AddressV = D3D11_TEXTURE_ADDRESS_CLAMP;
			samplerDesc.AddressW = D3D11_TEXTURE_ADDRESS_CLAMP;
			samplerDesc.ComparisonFunc = D3D11_COMPARISON_NEVER;
			samplerDesc.Filter = D3D11_FILTER_MIN_MAG_MIP_LINEAR;
			samplerDesc.MaxLOD = D3D11_FLOAT32_MAX;
			device->CreateSamplerState( &samplerDesc, &_samplerState );
		}

		if( _rasterState == NULL ) {
			D3D11_RASTERIZER_DESC rsdesc;
			memset( &rsdesc, 0, sizeof(rsdesc) );
			rsdesc.FillMode = D3D11_FILL_SOLID;
			rsdesc.CullMode = D3D11_CULL_NONE;
			//rsdesc.DepthClipEnable = TRUE;
			rsdesc.DepthClipEnable = FALSE;
			device->CreateRasterizerState( &rsdesc, &_rasterState );
		}

		if( _depthState == NULL ) {
			D3D11_DEPTH_STENCIL_DESC dsdesc;
			memset( &dsdesc, 0, sizeof(dsdesc) );
			//dsdesc.DepthEnable = TRUE;
			dsdesc.DepthEnable = FALSE;
			dsdesc.DepthWriteMask = D3D11_DEPTH_WRITE_MASK_ZERO;
			//dsdesc.DepthFunc = D3D11_COMPARISON_LESS_EQUAL;
			dsdesc.DepthFunc = D3D11_COMPARISON_ALWAYS;
			device->CreateDepthStencilState( &dsdesc, &_depthState );
		}

		if( _blendState == NULL ) {
			D3D11_BLEND_DESC bdesc;
			memset( &bdesc, 0, sizeof(bdesc) );
			bdesc.RenderTarget[0].BlendEnable = FALSE;
			bdesc.RenderTarget[0].RenderTargetWriteMask = 0xF;
			bdesc.RenderTarget[1].BlendEnable = FALSE;
			bdesc.RenderTarget[1].RenderTargetWriteMask = 0xF;
			device->CreateBlendState (&bdesc, &_blendState);
		}

		if( _pixelShader == NULL || _vertexShader == NULL ) {
			device->CreateVertexShader( g_VS, sizeof(g_VS), NULL, &_vertexShader );
			device->CreatePixelShader( g_PS, sizeof(g_PS), NULL, &_pixelShader );

			if( _vertexShader != NULL ) {
				if( _inputLayout == NULL ) {
					D3D11_INPUT_ELEMENT_DESC layout[] = {
						{ "POSITION", 0, DXGI_FORMAT_R32G32B32_FLOAT, 0, 0, D3D11_INPUT_PER_VERTEX_DATA, 0 },
						{ "TEXCOORD", 0, DXGI_FORMAT_R32G32_FLOAT, 0, 12, D3D11_INPUT_PER_VERTEX_DATA, 0 },
					};
					device->CreateInputLayout( layout, 2, g_VS, sizeof(g_VS), &_inputLayout );
				}
			}
		}

		if( _swapChain == NULL ) {
			_backBufferSize = rendererParams.backBufferSize;
			if( _backBufferSize.w <= 0 || _backBufferSize.h <= 0 ) {
				_backBufferSize = clientSize;
			}

			DXGI_SWAP_CHAIN_DESC scDesc;
			memset(&scDesc, 0, sizeof(scDesc));
			scDesc.BufferCount          = 2;
			scDesc.BufferDesc.Width     = _backBufferSize.w;
			scDesc.BufferDesc.Height    = _backBufferSize.h;
			scDesc.BufferDesc.Format    = DXGI_FORMAT_R8G8B8A8_UNORM;
			scDesc.BufferDesc.RefreshRate.Numerator   = 0;
			scDesc.BufferDesc.RefreshRate.Denominator = 1;
			scDesc.BufferUsage          = DXGI_USAGE_RENDER_TARGET_OUTPUT;
			scDesc.OutputWindow         = window->GetHWnd();
			scDesc.SampleDesc.Count     = 1;
			scDesc.SampleDesc.Quality   = 0;
			scDesc.Windowed             = TRUE;
			scDesc.Flags                = DXGI_SWAP_CHAIN_FLAG_ALLOW_MODE_SWITCH;
			scDesc.SwapEffect			= DXGI_SWAP_EFFECT_SEQUENTIAL;

			IDXGIDevice1* pDXGI = NULL;
			device->QueryInterface( __uuidof(IDXGIDevice1), (void**)&pDXGI );
			if( pDXGI != NULL ) {
				IDXGIAdapter* pAdapter = NULL;
				pDXGI->GetAdapter( &pAdapter );
				if( pAdapter != NULL ) {
					IDXGIFactory* pFactory = NULL;
					pAdapter->GetParent( __uuidof(IDXGIFactory), (void**)&pFactory );
					if( pFactory != NULL ) {
						pFactory->CreateSwapChain( device, &scDesc, &_swapChain );
						if( _swapChain != NULL ) { // Disable Alt + Enter to full screen for Oculus SDK.
							pFactory->MakeWindowAssociation( window->GetHWnd(), DXGI_MWA_NO_ALT_ENTER );
						}
						pFactory->Release();
					}
					pAdapter->Release();
				}
				pDXGI->Release();
			}

			if( _swapChain != NULL ) {
				HRESULT hr = _swapChain->GetBuffer( 0, __uuidof(ID3D11Texture2D), (void**)&_backBuffer );
				if( SUCCEEDED( hr ) ) {
					if( _backBuffer != NULL ) {
						device->CreateRenderTargetView( _backBuffer, NULL, &_backBufferRT );
					}
				}
			}
		}
	}

	ID3D11DeviceContext* ctx = NULL;
	device->GetImmediateContext( &ctx );
	if( ctx == NULL ) {
		return false;
	}

	if( _vertexBuffer == NULL ) {
		D3D11_BUFFER_DESC desc;
		memset( &desc, 0, sizeof(desc) );
		desc.Usage = D3D11_USAGE_DEFAULT;
		desc.ByteWidth = sizeof(_primitiveCache.vertices);
		desc.BindFlags = D3D11_BIND_VERTEX_BUFFER;
		desc.StructureByteStride = sizeof(RenderTexturePrimitiveCache::Vertex);
		device->CreateBuffer( &desc, NULL, &_vertexBuffer );
	}

	if( _vertexBuffer != NULL ) {
		if( _primitiveCache.Update( clientSize, rendererParams.viewportRect ) ) {
			ctx->UpdateSubresource( _vertexBuffer, 0, NULL, _primitiveCache.vertices, sizeof(_primitiveCache.vertices), 0 );
		}
	}

	ctx->Release();
	return true;
}

void DX11Renderer::_ShutdownDevice()
{
	SAFE_RELEASE(_textureView);
	SAFE_RELEASE(_backBufferRT);
	SAFE_RELEASE(_backBuffer);
	SAFE_RELEASE(_swapChain);
	SAFE_RELEASE(_samplerState);
	SAFE_RELEASE(_vertexBuffer);
	SAFE_RELEASE(_vertexShader);
	SAFE_RELEASE(_pixelShader);
	SAFE_RELEASE(_inputLayout);
	SAFE_RELEASE(_rasterState);
	SAFE_RELEASE(_blendState);
	SAFE_RELEASE(_depthState);

	{
		LockGuardRecursiveMutex lock(_mutex);
		SAFE_RELEASE(_renderTexture);
	}

	_isDeviceInitialized = false;
	memset( &_backBufferSize, 0, sizeof(_backBufferSize) );
	_primitiveCache.Clear();
}

} // namespace Windows
} // namespace OVRMirror
