// Copyright (c) 2014 Nora
// Released under the MIT license
// http://opensource.org/licenses/mit-license.php
#include "DX9Renderer.hpp"
#include "../Window.hpp"

namespace OVRMirror
{
namespace Windows
{

namespace
{

struct _Transform
{
	IDirect3DDevice9* _device;
	D3DTRANSFORMSTATETYPE _state;
	D3DMATRIX _matrix;

	_Transform( IDirect3DDevice9* device, D3DTRANSFORMSTATETYPE state, const D3DMATRIX& matrix )
		: _device( device )
		, _state( state )
	{
		memset( &_matrix, 0, sizeof(_matrix) );
		device->GetTransform( state, &_matrix );
		device->SetTransform( state, &matrix );
	}

	~_Transform()
	{
		_device->SetTransform( _state, &_matrix );
	}
};

struct _RenderState
{
	IDirect3DDevice9* _device;
	D3DRENDERSTATETYPE _state;
	DWORD _value;
	bool _isBackup;

	_RenderState( IDirect3DDevice9* device, D3DRENDERSTATETYPE state, DWORD value )
		: _device( device )
		, _state( state )
		, _value( 0 )
		, _isBackup( false )
	{
		device->GetRenderState( state, &_value );
		if( ( _isBackup = (_value != value) ) ) {
			device->SetRenderState( state, value );
		}
	}

	~_RenderState()
	{
		if( _isBackup ) {
			_device->SetRenderState( _state, _value );
		}
	}
};

struct _TextureStageState
{
	IDirect3DDevice9* _device;
	DWORD _stage;
	D3DTEXTURESTAGESTATETYPE _type;
	DWORD _value;
	bool _isBackup;

	_TextureStageState( IDirect3DDevice9* device, DWORD stage, D3DTEXTURESTAGESTATETYPE type, DWORD value )
		: _device( device )
		, _stage( stage )
		, _type( type )
		, _value( 0 )
		, _isBackup( false )
	{
		device->GetTextureStageState( stage, type, &_value );
		if( ( _isBackup = (_value != value) ) ) {
			device->SetTextureStageState( stage, type, value );
		}
	}

	~_TextureStageState()
	{
		if( _isBackup ) {
			_device->SetTextureStageState( _stage, _type, _value );
		}
	}
};


}

//--------------------------------------------------------------------------------------------------------------------------------

DX9Renderer::DX9Renderer()
	: _isDeviceInitialized( false )
	, _renderTexture( NULL )
	, _swapChain( NULL )
	, _backBuffer( NULL )
	, _vertexBuffer( NULL )
{
	memset( &_backBufferSize, 0, sizeof(_backBufferSize) );
}

DX9Renderer::~DX9Renderer()
{
	_ShutdownDevice();
}

void DX9Renderer::SetRenderTextureImmediately( void* renderTexture )
{
	LockGuardRecursiveMutex lock(_mutex);
	if( _renderTexture == renderTexture ) {
		return;
	}

	SAFE_RELEASE( _renderTexture );
	_renderTexture = static_cast< IDirect3DBaseTexture9* >( renderTexture );
	if( _renderTexture != NULL ) {
		_renderTexture->AddRef();
	}
}

void DX9Renderer::SetGraphicsDevice( void* device, int32_t eventType )
{
	switch( eventType ) {
	case kGfxDeviceEventInitialize:
	case kGfxDeviceEventAfterReset:
		// Nothing.
		break;
	case kGfxDeviceEventShutdown:
	case kGfxDeviceEventBeforeReset:
		_ShutdownDevice();
		break;
	}
}

void DX9Renderer::Render( const RendererParams& rendererParams )
{
	Window* window = static_cast< Window* >( rendererParams.window );
	if( window == nullptr ) {
		return;
	}

	bool isWindow = false;
	bool isMinimized = false;
	HWND hWnd = NULL;

	{
		LockGuardRecursiveMutex lock(window->mutex);
		isWindow = window->UnsafeCheckHWnd();
		isMinimized = window->UnsafeIsMinimized();
		hWnd = window->UnsafeGetHWnd();
	}

	if( !isWindow || rendererParams.openWindowTrigger == 0 ) {
		_ShutdownDevice();
		return;
	}

	if( isMinimized ) {
		return;
	}

	if( rendererParams.device == NULL ) {
		return;
	}

	if( !_PrepareDevice( rendererParams ) ) {
		return;
	}

	ISIZE32 clientSize = window->GetClientSizeImmediately();
	if( clientSize.w <= 0 || clientSize.h <= 0 ) {
		return;
	}

	IDirect3DDevice9* device = static_cast< IDirect3DDevice9* >( rendererParams.device );

	LPDIRECT3DSURFACE9 backupBackBuffer = NULL;
	if( FAILED( device->GetRenderTarget( 0, &backupBackBuffer ) ) || backupBackBuffer == NULL ) {
		return;
	}

	IDirect3DBaseTexture9* renderTexture = NULL;
	{
		LockGuardRecursiveMutex lock(_mutex);
		if( _renderTexture != NULL ) {
			renderTexture = _renderTexture;
			renderTexture->AddRef();
		}
	}

	if( _backBuffer != NULL && _vertexBuffer != NULL && renderTexture != NULL ) {
		device->SetRenderTarget( 0, _backBuffer );
		device->Clear( 0, NULL, D3DCLEAR_TARGET, D3DCOLOR_RGBA( 0, 0, 0, 0 ), 1.0f, 0 );

		{
			device->BeginScene();

			// Backup Texture.
			IDirect3DBaseTexture9* backupTexture = NULL;
			device->GetTexture( 0, &backupTexture );

			// Backup & Reset RenderState.
			_RenderState rs0( device, D3DRS_CULLMODE,			D3DCULL_NONE );
			_RenderState rs1( device, D3DRS_LIGHTING,			FALSE );
			_RenderState rs2( device, D3DRS_ALPHABLENDENABLE,	FALSE );
			_RenderState rs3( device, D3DRS_ALPHATESTENABLE,	FALSE );
			_RenderState rs4( device, D3DRS_ZFUNC,				D3DCMP_ALWAYS );
			_RenderState rs5( device, D3DRS_ZWRITEENABLE,		FALSE );
			_RenderState rs6( device, D3DRS_ZENABLE,			FALSE );

			const float identityMatrix[16] = {
				1, 0, 0, 0,
				0, 1, 0, 0,
				0, 0, 1, 0,
				0, 0, 0, 1,
			};

			// Backup & Reset Transforms.
			_Transform tr0( device, D3DTS_WORLD,		*(const D3DMATRIX*)identityMatrix );
			_Transform tr1( device, D3DTS_VIEW,			*(const D3DMATRIX*)identityMatrix );
			_Transform tr2( device, D3DTS_PROJECTION,	*(const D3DMATRIX*)identityMatrix );

			// Backup & Reset TextureStageState.
			_TextureStageState tss0( device, 0, D3DTSS_COLORARG1,	D3DTA_TEXTURE );
			_TextureStageState tss1( device, 0, D3DTSS_COLOROP,		D3DTOP_SELECTARG1 );
			_TextureStageState tss2( device, 0, D3DTSS_ALPHAARG1,	D3DTA_TEXTURE );
			_TextureStageState tss3( device, 0, D3DTSS_ALPHAOP,		D3DTOP_SELECTARG1 );
			_TextureStageState tss4( device, 1, D3DTSS_COLOROP,		D3DTOP_DISABLE );
			_TextureStageState tss5( device, 1, D3DTSS_ALPHAOP,		D3DTOP_DISABLE );

			// Reset Texture.
			device->SetTexture( 0, renderTexture );

			// Backup Viewport.
			D3DVIEWPORT9 backupViewport;
			memset( &backupViewport, 0, sizeof(backupViewport) );
			device->GetViewport( &backupViewport );

			// Reset Viewport.
			D3DVIEWPORT9 viewport;
			memset( &viewport, 0, sizeof(viewport) );
			viewport.Width = (DWORD)_backBufferSize.w;
			viewport.Height = (DWORD)_backBufferSize.h;
			viewport.MinZ = 0.0f;
			viewport.MaxZ = 1.0f;
			device->SetViewport( &viewport );

			// Backup StreamSource & FVF.
			IDirect3DVertexBuffer9* backupStreamData = NULL;
			UINT backupOffset = 0;
			UINT backupStride = 0;
			DWORD backupFVF = 0;
			device->GetStreamSource( 0, &backupStreamData, &backupOffset, &backupStride );
			device->GetFVF( &backupFVF );

			// Reset StreamSource & FVF.
			device->SetStreamSource( 0, _vertexBuffer, 0, sizeof(RenderTexturePrimitiveCache::Vertex) );
			device->SetFVF( D3DFVF_XYZ | D3DFVF_TEX1 );

			device->DrawPrimitive( D3DPT_TRIANGLESTRIP, 0, 2 );
			device->EndScene();
			_swapChain->Present( NULL, NULL, hWnd, NULL, 0 );

			// Restore StreamSource & FVF
			device->SetStreamSource( 0, backupStreamData, backupOffset, backupStride );
			device->SetFVF( backupFVF );
			SAFE_RELEASE( backupStreamData );

			// Restore Texture.
			device->SetTexture( 0, backupTexture );
			SAFE_RELEASE( backupTexture );

			// Restore Viewport.
			device->SetViewport( &backupViewport );
		}

		// Restore RenderTarget.
		device->SetRenderTarget( 0, backupBackBuffer );
	}

	{
		LockGuardRecursiveMutex lock(_mutex);
		SAFE_RELEASE( renderTexture );
	}

	SAFE_RELEASE( backupBackBuffer );
}

bool DX9Renderer::_PrepareDevice( const RendererParams& rendererParams )
{
	Window* window = static_cast< Window* >( rendererParams.window );
	if( window == nullptr ) {
		return false;
	}

	IDirect3DDevice9* device = static_cast< IDirect3DDevice9* >( rendererParams.device );
	if( device == NULL ) {
		return false;
	}

	ISIZE32 clientSize = window->GetClientSizeImmediately();
	if( clientSize.w <= 0 || clientSize.h <= 0 ) {
		return false;
	}

	const FRECT32& viewportRect = rendererParams.viewportRect;
	if( viewportRect.w <= 0.0f || viewportRect.h <= 0.0f ) {
		return false;
	}

	if( _swapChain != NULL ) {
		D3DPRESENT_PARAMETERS d3dpp;
		memset( &d3dpp, 0, sizeof(d3dpp) );
		if( SUCCEEDED( _swapChain->GetPresentParameters( &d3dpp ) ) ) {
			bool isResetSwapChain = false;
			HWND hWnd = d3dpp.hDeviceWindow;
			if( rendererParams.openWindowTrigger != -1 || hWnd != window->GetHWnd() || ::IsWindow( hWnd ) == FALSE ) {
				isResetSwapChain = true;
			}
			if( rendererParams.backBufferSize.w <= 0 ||
				rendererParams.backBufferSize.h <= 0 ) {
				if( d3dpp.BackBufferWidth != (UINT)clientSize.w || d3dpp.BackBufferHeight != (UINT)clientSize.h ) {
					isResetSwapChain = true;
				}
			}
			if( isResetSwapChain ) {
				SAFE_RELEASE( _swapChain );
				memset( &_backBufferSize, 0, sizeof(_backBufferSize) );
			}
		}
	}

	if( _swapChain == NULL ) {
		IDirect3DSwapChain9* swapChain = NULL;
		if( FAILED( device->GetSwapChain( 0, &swapChain ) ) || swapChain == NULL ) {
			return false;
		}

		D3DPRESENT_PARAMETERS d3dpp;
		memset( &d3dpp, 0, sizeof(d3dpp) );
		HRESULT hr = swapChain->GetPresentParameters( &d3dpp );
		swapChain->Release();
		if( FAILED( hr ) ) {
			return false;
		}

		_backBufferSize = rendererParams.backBufferSize;
		if( _backBufferSize.w <= 0 || _backBufferSize.h <= 0 ) {
			_backBufferSize = clientSize;
		}

		d3dpp.hDeviceWindow				= window->GetHWnd();
		d3dpp.Windowed					= TRUE;
		d3dpp.EnableAutoDepthStencil	= FALSE;
		d3dpp.AutoDepthStencilFormat	= D3DFMT_UNKNOWN;
		d3dpp.SwapEffect				= D3DSWAPEFFECT_DISCARD;
		d3dpp.PresentationInterval		= D3DPRESENT_INTERVAL_DEFAULT;
		d3dpp.MultiSampleType			= D3DMULTISAMPLE_NONE;
		d3dpp.MultiSampleQuality		= 0;
		d3dpp.Flags						= 0;
		d3dpp.BackBufferWidth			= _backBufferSize.w;
		d3dpp.BackBufferHeight			= _backBufferSize.h;
		d3dpp.BackBufferCount			= 0;

		// Memo: CreateAdditionalSwapChain() can't make full screen.
		if( FAILED( device->CreateAdditionalSwapChain( &d3dpp, &_swapChain ) ) || _swapChain == NULL ) {
			return false;
		}
	}

	if( _swapChain != NULL ) {
		if( FAILED( _swapChain->GetBackBuffer( 0, D3DBACKBUFFER_TYPE_MONO, &_backBuffer ) ) || _backBuffer == NULL ) {
			return false;
		}
	}

	if( _vertexBuffer == NULL ) {
		if( FAILED( device->CreateVertexBuffer( sizeof( _primitiveCache.vertices ),
			D3DUSAGE_WRITEONLY | D3DUSAGE_DYNAMIC, 0, D3DPOOL_DEFAULT, &_vertexBuffer, NULL ) )
			|| _vertexBuffer == NULL ) {
			return false;
		}
	}

	if( _vertexBuffer != NULL ) {
		if( _primitiveCache.Update( clientSize, rendererParams.viewportRect ) ) {
			void* vbPtr = nullptr;
			if( SUCCEEDED( _vertexBuffer->Lock( 0, 0, &vbPtr, D3DLOCK_DISCARD ) ) ) {
				if( vbPtr != nullptr ) {
					memcpy( vbPtr, _primitiveCache.vertices, sizeof( _primitiveCache.vertices ) );
				}
				_vertexBuffer->Unlock();
			}
		}
	}

	return true;
}

void DX9Renderer::_ShutdownDevice()
{
	SAFE_RELEASE(_vertexBuffer);
	SAFE_RELEASE(_backBuffer);
	SAFE_RELEASE(_swapChain);

	{
		LockGuardRecursiveMutex lock(_mutex);
		SAFE_RELEASE(_renderTexture);
	}

	_isDeviceInitialized = false;
	memset( &_backBufferSize, 0, sizeof(_backBufferSize) );
	_primitiveCache.Clear();
}

} // namespace Windows
} // namespace OVRMirror
