// Copyright (c) 2014 Nora
// Released under the MIT license
// http://opensource.org/licenses/mit-license.php
#include "OGLRenderer.hpp"
#include "../Window.hpp"

#ifdef _WIN32
#pragma comment(lib, "opengl32.lib")
#endif

namespace OVRMirror
{
namespace Windows
{

#define GL_TEXTURE0 0x84C0
#define GL_ARRAY_BUFFER 0x8892
#define GL_STATIC_DRAW 0x88E4
#define GL_ACTIVE_TEXTURE 0x84E0
#define GL_CLIENT_ACTIVE_TEXTURE 0x84E1

typedef void (WINAPI *_GLACTIVETEXTURE)( GLenum );
typedef void (WINAPI *_GLCLIENTACTIVETEXTURE)( GLenum );
typedef void (WINAPI *_GLGENBUFFERS)( GLsizei, GLuint* );
typedef void (WINAPI *_GLDELETEBUFFERS)( GLsizei, GLuint* );
typedef void (WINAPI *_GLBINDBUFFER)( GLenum, GLuint );
typedef void (WINAPI *_GLBUFFERDATA)( GLenum, ptrdiff_t, const GLvoid*, GLenum );
typedef void (WINAPI *_GLBUFFERSUBDATA)( GLenum, ptrdiff_t, ptrdiff_t, const GLvoid* );

OGLRenderer::OGLRenderer()
	: _isDeviceInitialized( false )
	, _hWnd( NULL )
	, _hDC( NULL )
	, _renderTexture( 0 )
	, _buffer( 0 )
{
	_glActiveTexture = ::wglGetProcAddress("glActiveTexture");
	_glClientActiveTexture = ::wglGetProcAddress("glClientActiveTexture");
	_glGenBuffers = ::wglGetProcAddress("glGenBuffers");
	_glDeleteBuffers = ::wglGetProcAddress("glDeleteBuffers");
	_glBindBuffer = ::wglGetProcAddress("glBindBuffer");
	_glBufferData = ::wglGetProcAddress("glBufferData");
	_glBufferSubData = ::wglGetProcAddress("glBufferSubData");
}

OGLRenderer::~OGLRenderer()
{
	_ShutdownDevice();
}

void OGLRenderer::SetRenderTextureImmediately( void* renderTexture )
{
	LockGuardRecursiveMutex lock(_mutex);
	_renderTexture = (GLuint)(size_t)renderTexture;
}

void OGLRenderer::SetGraphicsDevice( void* device, int32_t eventType )
{
	switch( eventType ) {
	case kGfxDeviceEventInitialize:
	case kGfxDeviceEventAfterReset:
		// Nothing.
		break;
	case kGfxDeviceEventShutdown:
	case kGfxDeviceEventBeforeReset:
		_ShutdownDevice();
		break;
	}
}

void OGLRenderer::Render( const RendererParams& rendererParams )
{
	Window* window = static_cast< Window* >( rendererParams.window );
	if( window == nullptr ) {
		return;
	}

	bool isWindow = false;
	bool isMinimized = false;

	{
		LockGuardRecursiveMutex lock(window->mutex);
		isWindow = window->UnsafeCheckHWnd();
		isMinimized = window->UnsafeIsMinimized();
	}

	if( !isWindow || rendererParams.openWindowTrigger == 0 ) {
		_ShutdownDevice();
		return;
	}

	if( isMinimized ) {
		return;
	}

	if( !_PrepareDevice( rendererParams ) ) {
		return;
	}

	GLuint renderTexture = 0;
	{
		LockGuardRecursiveMutex lock(_mutex);
		renderTexture = _renderTexture;
	}

	ISIZE32 clientSize = window->GetClientSizeImmediately();
	if( clientSize.w <= 0 || clientSize.h <= 0 ) {
		return;
	}

	HDC currentDC = ::wglGetCurrentDC();
	HGLRC currentContext = ::wglGetCurrentContext();
	if( currentDC == NULL || currentContext == NULL ) {
		return;
	}

	if( _hDC != NULL ) {
		::wglMakeCurrent( _hDC, currentContext );

		glClearColor( 0.0f, 0.0f, 0.0f, 0.0f );
		glClear( GL_COLOR_BUFFER_BIT );

		if( renderTexture != 0 ) {
			// Backup State.
			GLboolean backupCullFaceIsEnable = glIsEnabled( GL_CULL_FACE );
			GLboolean backupLightingIsEnable = glIsEnabled( GL_LIGHTING );
			GLboolean backupBlendIsEnable = glIsEnabled( GL_BLEND );
			GLboolean backupAlphaTestIsEnable = glIsEnabled( GL_ALPHA_TEST );
			GLboolean backupDepthTestIsEnable = glIsEnabled( GL_DEPTH_TEST );
			GLboolean backupTexture2DIsEnabled = glIsEnabled( GL_TEXTURE_2D );
			GLboolean backupDepthMask = GL_FALSE;
			glGetBooleanv( GL_DEPTH_WRITEMASK, &backupDepthMask );
			GLuint backupTexture2D = 0;
			glGetIntegerv( GL_TEXTURE_BINDING_2D, (GLint*)&backupTexture2D );
			GLint backupViewport[4] = { 0, 0, 0, 0 };
			glGetIntegerv( GL_VIEWPORT, backupViewport );
			GLint backupTextureEnvMode = 0;
			glGetTexEnviv( GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, &backupTextureEnvMode );
			GLfloat backupModelViewMatrix[16];
			GLfloat backupProjectionMatrix[16];
			memset( backupModelViewMatrix, 0, sizeof(backupModelViewMatrix) );
			memset( backupProjectionMatrix, 0, sizeof(backupProjectionMatrix) );
			glGetFloatv( GL_MODELVIEW_MATRIX, backupModelViewMatrix );
			glGetFloatv( GL_PROJECTION_MATRIX, backupProjectionMatrix );
			GLint backupMatrixMode = GL_MODELVIEW;
			glGetIntegerv( GL_MATRIX_MODE, &backupMatrixMode );
			GLint backupActiveTexture = GL_TEXTURE0;
			glGetIntegerv( GL_ACTIVE_TEXTURE, &backupActiveTexture );
			GLint backupClientActiveTexture = GL_TEXTURE0;
			glGetIntegerv( GL_CLIENT_ACTIVE_TEXTURE, &backupClientActiveTexture );

			// Reset State.
			glDisable( GL_CULL_FACE );
			glDisable( GL_LIGHTING );
			glDisable( GL_BLEND );
			glDisable( GL_ALPHA_TEST );
			glDisable( GL_DEPTH_TEST );
			glDepthMask( GL_FALSE );

			{
				// Reset Transforms.
				GLfloat identityMatrix[16] = {
					1, 0, 0, 0,
					0, 1, 0, 0,
					0, 0, 1, 0,
					0, 0, 0, 1,
				};
				glMatrixMode( GL_MODELVIEW );
				glLoadMatrixf( identityMatrix );
				glMatrixMode( GL_PROJECTION );
				glLoadMatrixf( identityMatrix );
			}

			if( _glClientActiveTexture != NULL && _glActiveTexture != NULL ) {
				(*(_GLCLIENTACTIVETEXTURE)_glClientActiveTexture)( GL_TEXTURE0 );
				(*(_GLACTIVETEXTURE)_glActiveTexture)( GL_TEXTURE0 );
			}

			glEnable( GL_TEXTURE_2D );
			glBindTexture( GL_TEXTURE_2D, renderTexture );
			glTexEnvi( GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE );

			glViewport( 0, 0, (GLsizei)clientSize.w, (GLsizei)clientSize.h );

			glEnableClientState( GL_VERTEX_ARRAY );
			glEnableClientState( GL_TEXTURE_COORD_ARRAY );
			glDisableClientState( GL_COLOR_ARRAY );

			if( _buffer != 0 ) {
				if( _glBindBuffer != NULL ) {
					(*(_GLBINDBUFFER)_glBindBuffer)( GL_ARRAY_BUFFER, _buffer );
					glVertexPointer( 3, GL_FLOAT, 0, 0 );
					glTexCoordPointer( 2, GL_FLOAT, 0, (const GLvoid*)(sizeof(float) * 3 * 4) );
					glDrawArrays( GL_TRIANGLE_STRIP, 0, 4 );
					(*(_GLBINDBUFFER)_glBindBuffer)( GL_ARRAY_BUFFER, 0 );
				}
			}

			glDisableClientState( GL_VERTEX_ARRAY );
			glDisableClientState( GL_TEXTURE_COORD_ARRAY );

			{
				// Restore Transforms.
				glMatrixMode( GL_MODELVIEW );
				glLoadMatrixf( backupModelViewMatrix );
				glMatrixMode( GL_PROJECTION );
				glLoadMatrixf( backupProjectionMatrix );

				glMatrixMode( backupMatrixMode );
			}

			// Restore Viewport.
			glViewport( backupViewport[0], backupViewport[1], (GLsizei)backupViewport[2], (GLsizei)backupViewport[3] );

			// Restore ClientActiveTexture & ActiveTexture.
			if( _glClientActiveTexture != NULL && _glActiveTexture != NULL ) {
				(*(_GLCLIENTACTIVETEXTURE)_glClientActiveTexture)( backupClientActiveTexture );
				(*(_GLACTIVETEXTURE)_glActiveTexture)( backupActiveTexture );
			}

			// Restore Texture.
			glBindTexture( GL_TEXTURE_2D, backupTexture2D );
			glTexEnvi( GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, backupTextureEnvMode );

			// Restore State.
			backupCullFaceIsEnable		? glEnable( GL_CULL_FACE )	: glDisable( GL_CULL_FACE );
			backupLightingIsEnable		? glEnable( GL_LIGHTING )	: glDisable( GL_LIGHTING );
			backupBlendIsEnable			? glEnable( GL_BLEND )		: glDisable( GL_BLEND );
			backupAlphaTestIsEnable		? glEnable( GL_ALPHA_TEST )	: glDisable( GL_ALPHA_TEST );
			backupDepthTestIsEnable		? glEnable( GL_DEPTH_TEST )	: glDisable( GL_DEPTH_TEST );
			backupTexture2DIsEnabled	? glEnable( GL_TEXTURE_2D )	: glDisable( GL_TEXTURE_2D ); // Memo: glDisable( GL_TEXTURE_2D ) at last.
			glDepthMask( backupDepthMask );
		}

		glFlush();
		::SwapBuffers( _hDC );
	}

	::wglMakeCurrent( currentDC, currentContext );
}

bool OGLRenderer::_PrepareDevice( const RendererParams& rendererParams )
{
	Window* window = static_cast< Window* >( rendererParams.window );
	if( window == nullptr ) {
		return false;
	}

	if( !rendererParams.deviceActivated ) {
		return false;
	}

	ISIZE32 clientSize = window->GetClientSizeImmediately();
	if( clientSize.w <= 0 || clientSize.h <= 0 ) {
		return false;
	}

	const FRECT32& viewportRect = rendererParams.viewportRect;
	if( viewportRect.w <= 0.0f || viewportRect.h <= 0.0f ) {
		return false;
	}

	HWND hWnd = window->GetHWnd();
	if( _hWnd != NULL ) {
		if( rendererParams.openWindowTrigger != -1 || _hWnd != hWnd || ::IsWindow( _hWnd ) == FALSE ) {
			_ShutdownDevice();
		}
	}

	if( !_isDeviceInitialized ) {
		if( _hDC == NULL ) {
			if( hWnd != NULL ) {
				::HDC currentDC = wglGetCurrentDC();
				::HGLRC currentContext = wglGetCurrentContext();
				if( currentDC != NULL && currentContext != NULL ) {
					_hDC = ::GetDC( hWnd );
					if( _hDC != NULL ) {
						int pixelFormat = ::GetPixelFormat( currentDC );
						::PIXELFORMATDESCRIPTOR pfd;
						memset( &pfd, 0, sizeof(pfd) );
						::DescribePixelFormat( currentDC, pixelFormat, sizeof(pfd), &pfd );
						::SetPixelFormat( _hDC, pixelFormat, &pfd );
					}
				}
			}
			if( _hDC == NULL ) {
				return false;
			}
		}

		_isDeviceInitialized = true;

		if( _buffer == 0 ) {
			if( _glGenBuffers != NULL ) {
				(*(_GLGENBUFFERS)_glGenBuffers)( 1, &_buffer );
			}
			if( _buffer != 0 ) {
				if( _glBindBuffer != NULL && _glBufferData != NULL ) {
					(*(_GLBINDBUFFER)_glBindBuffer)( GL_ARRAY_BUFFER, _buffer );
					(*(_GLBUFFERDATA)_glBufferData)( GL_ARRAY_BUFFER, sizeof(float) * (3 + 2) * 4, NULL, GL_STATIC_DRAW );
					(*(_GLBINDBUFFER)_glBindBuffer)( GL_ARRAY_BUFFER, 0 );
				}
			}
		}
	}

	if( _buffer != 0 ) {
		if( _primitiveCache.Update( clientSize, rendererParams.viewportRect ) ) {
			if( _glBindBuffer != NULL && _glBufferSubData != NULL ) {
				float vertices[(3 + 2) * 4];
				for( int i = 0; i != 4; ++i ) {
					memcpy( &(vertices[3 * i]), _primitiveCache.vertices[i].pos, sizeof(float) * 3 );
					memcpy( &(vertices[3 * 4 + 2 * i]), _primitiveCache.vertices[i].uv, sizeof(float) * 2 );
				}

				(*(_GLBINDBUFFER)_glBindBuffer)( GL_ARRAY_BUFFER, _buffer );
				(*(_GLBUFFERSUBDATA)_glBufferSubData)( GL_ARRAY_BUFFER, 0, sizeof(vertices), vertices );
				(*(_GLBINDBUFFER)_glBindBuffer)( GL_ARRAY_BUFFER, 0 );
			}
		}
	}

	return true;
}

void OGLRenderer::_ShutdownDevice()
{
	{
		LockGuardRecursiveMutex lock(_mutex);
		_renderTexture = 0;
	}

	if( _buffer != 0 ) {
		if( _glDeleteBuffers != NULL ) {
			(*(_GLDELETEBUFFERS)_glDeleteBuffers)( 1, &_buffer );
		}
		_buffer = 0;
	}

	if( _hDC != NULL ) {
		::ReleaseDC( _hWnd, _hDC );
		_hDC = NULL;
	}

	_primitiveCache.Clear();

	_hWnd = NULL;
	_isDeviceInitialized = false;
}

} // namespace Windows
} // namespace OVRMirror
