// Copyright (c) 2014 Nora
// Released under the MIT license
// http://opensource.org/licenses/mit-license.php
#include "Window.hpp"
#include <map>

#define _BIND_ROOTWINDOW

namespace OVRMirror
{
namespace Windows
{

//----------------------------------------------------------------------------------------------------------------------------------------

#define _WS_OVRMIRROR_WINDOWED (WS_OVERLAPPED     | \
								WS_CAPTION        | \
								WS_THICKFRAME)

static const wchar_t* _className = L"OVRMirror";

static inline const wchar_t* _GetWindowTitle( int32_t deviceType )
{
	switch( deviceType ) {
	case kGfxRendererOpenGL:
		return L"OVRMirror <OpenGL>";
	case kGfxRendererD3D9:
		return L"OVRMirror <DX9>";
	case kGfxRendererD3D11:
		return L"OVRMirror <DX11>";
	}

	return L"OVRMirror";
}

static HWND _rootWindow = NULL;
static bool _isActiveRootWindow = false;

//----------------------------------------------------------------------------------------------------------------------------------------

static inline void _SetActiveOculusWindow( HWND hWnd )
{
	if( ::IsIconic( hWnd ) != FALSE ) { // Restore Extend.
		::ShowWindow( hWnd, SW_RESTORE );
	}
	::SetActiveWindow( hWnd );
	::SetForegroundWindow( hWnd );
	::SetFocus( hWnd );
	::UpdateWindow( hWnd );
}

//----------------------------------------------------------------------------------------------------------------------------------------

Window::Window()
	: _hWnd( NULL )
	, _deviceType( -1 )
{
}

Window::~Window()
{
	_DestroyWindow();
}

void Window::SetDeviceType( int32_t deviceType )
{
	LockGuardRecursiveMutex lock(mutex);
	if( _deviceType != deviceType ) {
		_deviceType = deviceType;
		if( deviceType != -1 ) {
			if( _hWnd != NULL ) {
				::SetWindowTextW( _hWnd, _GetWindowTitle( deviceType ) );
			}
		}
	}
}

void Window::_CreateWindow( int32_t deviceType, const ISIZE32& clientSize, bool isFullscreen, bool isMaximized )
{
	HWND hWndParent = NULL; // Don't set.(Can't boot mirror mode.)
	const HINSTANCE hInstance = ::GetModuleHandleW( NULL );
	const DWORD dwStyle = (isFullscreen ? WS_POPUP : _WS_OVRMIRROR_WINDOWED);

	INT x = CW_USEDEFAULT;
	INT y = CW_USEDEFAULT;
	INT w = CW_USEDEFAULT;
	INT h = CW_USEDEFAULT;
	if( !isFullscreen && !isMaximized ) {
		if( clientSize.w > 0 && clientSize.h > 0 ) {
			RECT rc = { 0, 0, (INT)clientSize.w, (INT)clientSize.h };
			if( ::AdjustWindowRect( &rc, dwStyle, FALSE ) != FALSE ) {
				w = (INT)(rc.right - rc.left);
				h = (INT)(rc.bottom - rc.top);
			}
		}
	} else if( !isFullscreen && isMaximized ) {
		::RECT rc = { 0, 0, 0, 0 };
		::SystemParametersInfoW( SPI_GETWORKAREA, 0, &rc, 0 );
		x = (INT)rc.left;
		y = (INT)rc.top;
		w = (INT)(rc.right - rc.left);
		h = (INT)(rc.bottom - rc.top);
	}

	_RegisterClass();

	LockGuardRecursiveMutex lock(mutex);

	_deviceType = deviceType;

	if( _hWnd != NULL ) { // Memo: Not use UnsafeCheckHWnd()
		::DestroyWindow( _hWnd );
		_hWnd = NULL;
	}

	_hWnd = ::CreateWindowW( _className , _GetWindowTitle( deviceType ), dwStyle,
				x, y, w, h,
				hWndParent, NULL, hInstance, NULL );

	if( _hWnd != NULL ) {
		::SetWindowPos( _hWnd, HWND_TOPMOST, 0, 0, 0, 0, SWP_NOMOVE | SWP_NOSIZE | SWP_NOACTIVATE );
		::ShowWindow( _hWnd, SW_SHOWNA );
		::UpdateWindow( _hWnd );

		_SetActiveOculusWindow( _rootWindow );
	}
}

void Window::_DestroyWindow()
{
	LockGuardRecursiveMutex lock(mutex);
	if( _hWnd != NULL ) { // Memo: Not use UnsafeCheckHWnd()
		::DestroyWindow( _hWnd );
		_hWnd = NULL;
	}
}

void Window::_ShowWindow()
{
	LockGuardRecursiveMutex lock(mutex);
	if( !UnsafeCheckHWnd() ) {
		return;
	}

	if( UnsafeIsMinimized() ) {
		::ShowWindow( _hWnd, SW_RESTORE | SW_SHOWNA );
	} else {
		::ShowWindow( _hWnd, SW_SHOWNA );
	}

	::UpdateWindow( _hWnd );
}

void Window::_AdjustWindowSize( const ISIZE32& clientSize )
{
	LockGuardRecursiveMutex lock(mutex);
	if( !UnsafeCheckHWnd() ) {
		return;
	}
	if( UnsafeIsMaximized() || UnsafeIsMinimized() ) {
		return;
	}

	RECT rcWin = { 0, 0, 0, 0 };
	RECT rcCli = { 0, 0, 0, 0 };
	if( ::GetWindowRect( _hWnd, &rcWin ) == FALSE ) {
		return;
	}
	if( ::GetClientRect( _hWnd, &rcCli ) == FALSE ) {
		return;
	}

	LONG padW = (rcWin.right - rcWin.left) - (rcCli.right - rcCli.left);
	LONG padH = (rcWin.bottom - rcWin.top) - (rcCli.bottom - rcCli.top);

	::SetWindowPos( _hWnd, NULL, 0, 0,
		(int)padW + (int)clientSize.w,
		(int)padH + (int)clientSize.h,
		SWP_NOMOVE | SWP_NOZORDER );
}

void Window::_ChangeWindowStyle_Fullscreen( bool isFullscreen )
{
	LockGuardRecursiveMutex lock(mutex);
	if( !UnsafeCheckHWnd() ) {
		return;
	}

	DWORD dwStyle = 0;
	if( isFullscreen ) {
		dwStyle = WS_POPUP;
	} else {
		dwStyle = _WS_OVRMIRROR_WINDOWED;
	}

	if( dwStyle != (DWORD)::GetWindowLongW( _hWnd, GWL_STYLE ) ) {
		::SetWindowLongW( _hWnd, GWL_STYLE, (LONG)dwStyle );
		::UpdateWindow( _hWnd );
	}
}

int32_t Window::GetWindowStateImmediately()
{
	LockGuardRecursiveMutex lock(mutex);
	if( !UnsafeCheckHWnd() ) {
		return 0;
	}

	int32_t r = 0x0001;

	if( UnsafeIsMaximized() ) {
		r |= 0x0002;
	}
	if( UnsafeIsMinimized() ) {
		r |= 0x0004;
	}

	POINT pt = { 0, 0 };
	if( ::GetCursorPos( &pt ) != FALSE ) {
		HWND hWnd = ::WindowFromPoint( pt );
		if( hWnd != NULL && hWnd == _hWnd ) {
			r |= 0x0008;
		}
	}

	return r;
}

//----------------------------------------------------------------------------------------------------------------------------------------

void Window::_RegisterClass()
{
	HINSTANCE hInstance = ::GetModuleHandleW( NULL );

	WNDCLASSW wc;
	memset( &wc, 0, sizeof(wc) );
	if( ::GetClassInfoW( hInstance, _className, &wc ) != FALSE ) { // If already RegisterClassW(), return TRUE.
		return;
	}

	memset( &wc, 0, sizeof(wc) );
	wc.hInstance     = hInstance;
	wc.hCursor       = LoadCursor( NULL , IDC_ARROW );
	wc.hIcon         = _GetRootWindowIcon( false );
	if( wc.hIcon == NULL ) {
		wc.hIcon     = LoadIcon( NULL , IDI_APPLICATION );
	}
	wc.hbrBackground = NULL;
	wc.lpszClassName = _className;
	wc.lpszMenuName  = NULL;
	wc.lpfnWndProc   = _WndProc;
	wc.cbWndExtra    = 0;
	wc.cbClsExtra    = 0;

	(void)::RegisterClassW( &wc );
}

LRESULT CALLBACK Window::_WndProc( HWND hWnd, UINT msg, WPARAM wParam, LPARAM lParam )
{
	switch( msg ) {
	case WM_CREATE:
		_rootWindow = _GetRootWindow();
		_isActiveRootWindow = false;
		::SetTimer( hWnd, 100, 10, NULL );
		break;
	case WM_TIMER:
		if( wParam == 100 ) {
			if( ::IsIconic( _rootWindow ) ) {
				::PostMessageW( hWnd, WM_USER + 100, 0, 0 );
			}
		}
		break;
	case WM_MOVE:
		::PostMessageW( hWnd, WM_USER + 100, 0, 0 );
		break;
	case WM_USER + 100:
		_isActiveRootWindow = true;
		_SetActiveOculusWindow( _rootWindow );
		break;
	case WM_NCACTIVATE:
		if( _isActiveRootWindow ) {
			_isActiveRootWindow = false;
			::DefWindowProc( hWnd, msg, TRUE, 0 );
			return TRUE;
		}
		break;
	case WM_ACTIVATE:
		if( HIWORD(wParam) == FALSE ) {
			if( LOWORD( wParam ) == WA_ACTIVE || LOWORD( wParam ) == WA_CLICKACTIVE ) {
				::PostMessageW( hWnd, WM_USER + 100, 0, 0 );
			}
		}
		break;
	case WM_CLOSE:
		::DestroyWindow( hWnd );
		return 0;
	case WM_DESTROY:
		::KillTimer( hWnd, 100 );
		_SetActiveOculusWindow( _rootWindow );
		_rootWindow = NULL;
		_isActiveRootWindow = false;
		break;
	}
	
	return ::DefWindowProc( hWnd, msg, wParam, lParam );
}

static bool _IsOVRMirror( HWND hWnd )
{
	wchar_t classNameBuf[256] = { 0 };
	int len = ::GetClassNameW( hWnd, classNameBuf, 256 );
	if( len > 0 ) {
		int classNameLen = (int)wcslen( _className );
		if( classNameLen <= len &&
			memcmp( _className, classNameBuf, classNameLen * sizeof(wchar_t) ) == 0 ) {
			return true;
		}
	}

	return false;
}

HWND Window::_GetRootWindow()
{
	DWORD currentProcessID = ::GetCurrentProcessId();
	HWND hWnd = GetTopWindow( GetDesktopWindow() );
	while( hWnd != NULL ) {
		DWORD dwProcessId = 0;
		GetWindowThreadProcessId( hWnd, &dwProcessId );
		if( dwProcessId == currentProcessID ) {
			if( !_IsOVRMirror( hWnd ) ) {
				if( ::GetParent( hWnd ) == NULL ) {
					return hWnd;
				}
			}
		}

		hWnd = GetNextWindow(hWnd, GW_HWNDNEXT);
	}

	return NULL;
}

HICON Window::_GetRootWindowIcon( bool isSmall )
{
	HWND hWnd = _GetRootWindow();
	if( hWnd != NULL ) {
#ifdef _WIN64
		HICON hIcon = (HICON)GetClassLongPtr(hWnd, isSmall ? GCLP_HICONSM : GCLP_HICON);
#else
		HICON hIcon = (HICON)GetClassLong(hWnd, isSmall ? GCL_HICONSM : GCL_HICON);
#endif
		return hIcon;
	}

	return NULL;
}

} // namespace Windows
} // namespace OVRMirror
